import logging
import random
import string
from collections.abc import Callable, Generator
from typing import Any, TypeVar

from pytest import fixture

from databricks.sdk import WorkspaceClient, AccountClient
from databricks.sdk.config import Config
from databricks.sdk.errors import DatabricksError

_LOG = logging.getLogger(__name__)


@fixture
def make_random():
    """
    Fixture to generate random strings.

    This fixture provides a function to generate random strings of a specified length.
    The generated strings are created using a character set consisting of uppercase letters,
    lowercase letters, and digits.

    To generate a random string with default length of 16 characters:

    ```python
    random_string = make_random()
    assert len(random_string) == 16
    ```

    To generate a random string with a specified length:

    ```python
    random_string = make_random(k=8)
    assert len(random_string) == 8
    ```
    """

    def inner(k=16) -> str:
        """
        Generate a random string.

        Parameters:
        -----------
        k : int, optional
            Length of the random string (default is 16).

        Returns:
        --------
        str:
            A randomly generated string.
        """
        charset = string.ascii_uppercase + string.ascii_lowercase + string.digits
        return "".join(random.choices(charset, k=int(k)))

    return inner


T = TypeVar("T")


def factory(
    name: str,
    create: Callable[..., T],
    remove: Callable[[T], None],
) -> Generator[Callable[..., T], None, None]:
    """
    Factory function for creating fixtures.

    This function creates a fixture for managing resources (e.g., secret scopes) within test functions.
    The provided ``create`` function is used to create a resource, and the provided ``remove`` function
    is used to remove the resource after the test is complete.

    Parameters:
    -----------
    name : str
        The name of the resource (e.g., "secret scope").
    create : function
        A function to create the resource.
    remove : function
        A function to remove the resource.

    Returns:
    --------
    function:
        A function to create and manage the resource fixture.

    Usage Example:
    --------------
    To create a fixture for managing secret scopes:

    .. code-block:: python

       @pytest.fixture
       def make_secret_scope(ws, make_random):
           def create(**kwargs):
               name = f"sdk-{make_random(4)}"
               ws.secrets.create_scope(name, **kwargs)
               return name

           def cleanup(scope):
               ws.secrets.delete_scope(scope)

           yield from factory("secret scope", create, cleanup)
    """
    cleanup: list[T] = []

    def inner(**kwargs: Any) -> T:
        out = create(**kwargs)
        _LOG.debug(f"added {name} fixture: {out}")
        cleanup.append(out)
        return out

    yield inner
    _LOG.debug(f"clearing {len(cleanup)} {name} fixtures")
    for some in cleanup:
        try:
            _LOG.debug(f"removing {name} fixture: {some}")
            remove(some)
        except DatabricksError as e:
            _LOG.debug(f"ignoring error while {name} {some} teardown: {e}")


@fixture
def product_info():
    return None, None


@fixture
def ws(debug_env: dict[str, str], product_info: tuple[str, str]) -> WorkspaceClient:
    """
    Create and provide a Databricks WorkspaceClient object.

    This fixture initializes a Databricks WorkspaceClient object, which can be used
    to interact with the Databricks workspace API. The created instance of WorkspaceClient
    is shared across all test functions within the test session.

    See [detailed documentation](https://databricks-sdk-py.readthedocs.io/en/latest/authentication.html) for the list
    of environment variables that can be used to authenticate the WorkspaceClient.

    In your test functions, include this fixture as an argument to use the WorkspaceClient:

    ```python
    def test_workspace_operations(ws):
        clusters = ws.clusters.list_clusters()
        assert len(clusters) >= 0
    ```
    """
    product_name, product_version = product_info
    return WorkspaceClient(
        host=debug_env["DATABRICKS_HOST"],
        auth_type=debug_env.get("DATABRICKS_AUTH_TYPE"),
        token=debug_env.get("DATABRICKS_TOKEN"),
        username=debug_env.get("DATABRICKS_USERNAME"),
        password=debug_env.get("DATABRICKS_PASSWORD"),
        client_id=debug_env.get("DATABRICKS_CLIENT_ID"),
        client_secret=debug_env.get("DATABRICKS_CLIENT_SECRET"),
        debug_truncate_bytes=debug_env.get("DATABRICKS_DEBUG_TRUNCATE_BYTES"),  # type: ignore
        debug_headers=debug_env.get("DATABRICKS_DEBUG_HEADERS"),  # type: ignore
        azure_client_id=debug_env.get("ARM_CLIENT_ID"),
        azure_tenant_id=debug_env.get("ARM_TENANT_ID"),
        azure_client_secret=debug_env.get("ARM_CLIENT_SECRET"),
        cluster_id=debug_env.get("DATABRICKS_CLUSTER_ID"),
        product=product_name,
        product_version=product_version,
    )


@fixture
def acc(debug_env: dict[str, str], product_info: tuple[str, str], env_or_skip) -> AccountClient:
    """
    Create and provide a Databricks AccountClient object.

    This fixture initializes a Databricks AccountClient object, which can be used
    to interact with the Databricks account API. The created instance of AccountClient
    is shared across all test functions within the test session.

    Requires `DATABRICKS_ACCOUNT_ID` environment variable to be set. If `DATABRICKS_HOST`
    points to a workspace host, the fixture would automatically determine the account host
    from it.

    See [detailed documentation](https://databricks-sdk-py.readthedocs.io/en/latest/authentication.html) for the list
    of environment variables that can be used to authenticate the AccountClient.

    In your test functions, include this fixture as an argument to use the AccountClient:

    ```python
    def test_listing_workspaces(acc):
        workspaces = acc.workspaces.list()
        assert len(workspaces) >= 1
    ```
    """
    product_name, product_version = product_info
    config = Config(
        host=debug_env["DATABRICKS_HOST"],
        account_id=env_or_skip("DATABRICKS_ACCOUNT_ID"),
        auth_type=debug_env.get("DATABRICKS_AUTH_TYPE"),
        username=debug_env.get("DATABRICKS_USERNAME"),
        password=debug_env.get("DATABRICKS_PASSWORD"),
        client_id=debug_env.get("DATABRICKS_CLIENT_ID"),
        client_secret=debug_env.get("DATABRICKS_CLIENT_SECRET"),
        debug_truncate_bytes=debug_env.get("DATABRICKS_DEBUG_TRUNCATE_BYTES"),  # type: ignore
        debug_headers=debug_env.get("DATABRICKS_DEBUG_HEADERS"),  # type: ignore
        azure_client_id=debug_env.get("ARM_CLIENT_ID"),
        azure_tenant_id=debug_env.get("ARM_TENANT_ID"),
        azure_client_secret=debug_env.get("ARM_CLIENT_SECRET"),
        product=product_name,
        product_version=product_version,
    )
    config.host = config.environment.deployment_url('accounts')
    return AccountClient(config=config)


@fixture
def log_workspace_link(ws):
    """Returns a function to log a workspace link."""

    def inner(name: str, path: str, *, anchor: bool = True):
        a = '#' if anchor else ''
        url = f'https://{ws.config.hostname}/{a}{path}'
        _LOG.info(f'Created {name}: {url}')

    return inner


@fixture
def log_account_link(acc):
    """Returns a function to log an account link."""

    def inner(name: str, path: str, *, anchor: bool = False):
        a = '#' if anchor else ''
        url = f'https://{acc.config.hostname}/{a}{path}'
        _LOG.info(f'Created {name}: {url}')

    return inner



import json
from collections.abc import Callable, Generator

from pytest import fixture
from databricks.sdk.service.sql import LegacyQuery

from databricks.labs.pytester.fixtures.baseline import factory


@fixture
def make_query(
    ws,
    make_table,
    make_random,
    log_workspace_link,
    watchdog_remove_after,
) -> Generator[Callable[..., LegacyQuery], None, None]:
    """
    Create a query and remove it after the test is done. Returns the `databricks.sdk.service.sql.LegacyQuery` object.

    Keyword Arguments:
    - `sql_query`: The query to be stored. Default is `SELECT * FROM <newly created random table>`.

    Usage:
    ```python
    from databricks.sdk.service.sql import PermissionLevel

    def test_permissions_for_redash(
        make_user,
        make_query,
        make_query_permissions,
    ):
        user = make_user()
        query = make_query()
        make_query_permissions(
            object_id=query.id,
            permission_level=PermissionLevel.CAN_EDIT,
            user_name=user.display_name,
        )
    ```
    """

    def create(*, sql_query: str | None = None, **kwargs) -> LegacyQuery:
        if sql_query is None:
            table = make_table()
            sql_query = f"SELECT * FROM {table.catalog_name}.{table.schema_name}.{table.name}"
        # add RemoveAfter tag for watchdog
        remove_after_tag = json.dumps({"key": "RemoveAfter", "value": watchdog_remove_after})
        tags: list[str] | None = kwargs.get('tags', None)
        if tags:
            tags.append(remove_after_tag)
        else:
            tags = [remove_after_tag]
        query_name = f"dummy_query_Q{make_random(4)}"
        query = ws.queries_legacy.create(
            name=query_name,
            description="Test query",
            query=sql_query,
            tags=tags,
        )
        log_workspace_link(f"{query_name} query", f'sql/editor/{query.id}')
        return query

    def remove(query: LegacyQuery):
        ws.queries_legacy.delete(query_id=query.id)

    yield from factory("query", create, remove)



import logging

from pytest import fixture, skip
from databricks.sdk import WorkspaceClient

logger = logging.getLogger(__name__)


@fixture
def spark(ws: WorkspaceClient):
    """
    Get Databricks Connect Spark session. Requires `databricks-connect` package to be installed.

    To enable serverless set the local environment variable `DATABRICKS_SERVERLESS_COMPUTE_ID` to `"auto"`.
    If this environment variable is set, Databricks Connect ignores the cluster_id.
    If `DATABRICKS_SERVERLESS_COMPUTE_ID` is set to a specific serverless cluster ID, that cluster will be used instead.
    However, this is not recommended, as serverless clusters are ephemeral by design.
    See more details [here](https://docs.databricks.com/en/dev-tools/databricks-connect/cluster-config.html#configure-a-connection-to-serverless-compute).

    Usage:
    ```python
    def test_databricks_connect(spark):
        rows = spark.sql("SELECT 1").collect()
        assert rows[0][0] == 1
    ```
    """
    cluster_id = ws.config.cluster_id
    serverless_cluster_id = ws.config.serverless_compute_id

    if not serverless_cluster_id:
        ensure_cluster_is_running(cluster_id, ws)

    if serverless_cluster_id and serverless_cluster_id != "auto":
        ensure_cluster_is_running(serverless_cluster_id, ws)

    try:
        # pylint: disable-next=import-outside-toplevel
        from databricks.connect import (  # type: ignore[import-untyped]
            DatabricksSession,
        )

        if serverless_cluster_id:
            logging.debug(f"Using serverless cluster id '{serverless_cluster_id}'")
            return DatabricksSession.builder.serverless(True).getOrCreate()

        logging.debug(f"Using cluster id '{cluster_id}'")
        return DatabricksSession.builder.sdkConfig(ws.config).getOrCreate()
    except ImportError:
        skip("Please run `pip install databricks-connect`")
        return None


def ensure_cluster_is_running(cluster_id: str, ws: WorkspaceClient) -> None:
    if not cluster_id:
        skip("No cluster_id found in the environment")
    ws.clusters.ensure_cluster_is_running(cluster_id)



import logging
from collections.abc import Callable, Generator
from unittest.mock import Mock

from pytest import fixture
from databricks.sdk.errors import BadRequest
from databricks.sdk.service._internal import Wait
from databricks.sdk.service.serving import (
    EndpointCoreConfigInput,
    EndpointPendingConfig,
    EndpointTag,
    ServedModelInput,
    ServedModelOutput,
    ServingEndpointDetailed,
)
from databricks.sdk.service.ml import CreateExperimentResponse, ModelDatabricks, ModelTag

from databricks.labs.pytester.fixtures.baseline import factory


logger = logging.getLogger(__name__)


@fixture
def make_experiment(
    ws,
    make_random,
    make_directory,
    log_workspace_link,
    watchdog_purge_suffix,
) -> Generator[Callable[..., CreateExperimentResponse], None, None]:
    """
    Returns a function to create Databricks Experiments and clean them up after the test.
    The function returns a `databricks.sdk.service.ml.CreateExperimentResponse` object.

    Keyword arguments:
    * `path` (str, optional): The path of the experiment. Defaults to `dummy-*` experiment in current user's home folder.
    * `experiment_name` (str, optional): The name of the experiment. Defaults to `dummy-*`.

    Usage:
    ```python
    from databricks.sdk.service.iam import PermissionLevel

    def test_experiments(make_group, make_experiment, make_experiment_permissions):
        group = make_group()
        experiment = make_experiment()
        make_experiment_permissions(
            object_id=experiment.experiment_id,
            permission_level=PermissionLevel.CAN_MANAGE,
            group_name=group.display_name,
        )
    ```
    """

    def create(
        *,
        path: str | None = None,
        experiment_name: str | None = None,
        **kwargs,
    ) -> CreateExperimentResponse:
        folder = make_directory(path=path)
        if experiment_name is None:
            # The purge suffix is needed here as well, just in case the path was supplied.
            experiment_name = f"dummy-{make_random(4)}-{watchdog_purge_suffix}"
        experiment = ws.experiments.create_experiment(name=f"{folder}/{experiment_name}", **kwargs)
        log_workspace_link(f'{experiment_name} experiment', f'ml/experiments/{experiment.experiment_id}', anchor=False)
        return experiment

    yield from factory("experiment", create, lambda item: ws.experiments.delete_experiment(item.experiment_id))


@fixture
def make_model(ws, make_random, watchdog_remove_after) -> Generator[Callable[..., ModelDatabricks], None, None]:
    """
    Returns a function to create Databricks Models and clean them up after the test.
    The function returns a `databricks.sdk.service.ml.GetModelResponse` object.

    Keyword arguments:
    * `model_name` (str, optional): The name of the model. Defaults to `dummy-*`.

    Usage:
    ```python
    from databricks.sdk.service.iam import PermissionLevel

    def test_models(make_group, make_model, make_registered_model_permissions):
        group = make_group()
        model = make_model()
        make_registered_model_permissions(
            object_id=model.id,
            permission_level=PermissionLevel.CAN_MANAGE,
            group_name=group.display_name,
        )
    ```
    """

    def create(*, model_name: str | None = None, **kwargs) -> ModelDatabricks:
        if model_name is None:
            model_name = f"dummy-{make_random(4)}"
        remove_after_tag = ModelTag(key="RemoveAfter", value=watchdog_remove_after)
        if 'tags' not in kwargs:
            kwargs["tags"] = [remove_after_tag]
        else:
            kwargs["tags"].append(remove_after_tag)
        created_model = ws.model_registry.create_model(model_name, **kwargs)
        model = ws.model_registry.get_model(created_model.registered_model.name)
        assert model.registered_model_databricks is not None
        return model.registered_model_databricks

    yield from factory("model", create, lambda item: ws.model_registry.delete_model(item.id))


@fixture
def make_serving_endpoint(ws, make_random, watchdog_remove_after):
    """
    Returns a function to create Databricks Serving Endpoints and clean them up after the test.
    The function returns a `databricks.sdk.service.serving.ServingEndpointDetailed` object.

    Under the covers, this fixture also creates a model to serve on a small workload size.

    Keyword arguments:
    * `endpoint_name` (str, optional): The name of the endpoint. Defaults to `dummy-*`.
    * `model_name` (str, optional): The name of the model to serve on the endpoint.
        Defaults to system model `system.ai.llama_v3_2_1b_instruct`.
    * `model_version` (str, optional): The model version to serve. If None, tries to get the latest version for
        workspace local models. Otherwise, defaults to version `1`.

    Usage:
    ```python
    def test_endpoints(make_group, make_serving_endpoint, make_serving_endpoint_permissions):
        group = make_group()
        endpoint = make_serving_endpoint()
        make_serving_endpoint_permissions(
            object_id=endpoint.response.id,
            permission_level=PermissionLevel.CAN_QUERY,
            group_name=group.display_name,
        )
    ```
    """

    def create(
        *,
        endpoint_name: str | None = None,
        model_name: str | None = None,
        model_version: str | None = None,
    ) -> Wait[ServingEndpointDetailed]:
        endpoint_name = endpoint_name or make_random(4)
        model_name = model_name or "system.ai.llama_v3_2_1b_instruct"
        if not model_version and "." not in model_name:  # The period in the name signals it is NOT workspace local
            try:
                model_version = ws.model_registry.get_latest_versions(model_name).version
            except BadRequest as e:
                logger.warning(
                    f"Cannot get latest version for model: {model_name}. Fallback to version '1'.", exc_info=e
                )
        model_version = model_version or "1"
        tags = [EndpointTag(key="RemoveAfter", value=watchdog_remove_after)]
        served_model_input = ServedModelInput(
            model_name=model_name,
            model_version=model_version,
            scale_to_zero_enabled=True,
            workload_size="Small",
        )
        endpoint = ws.serving_endpoints.create(
            endpoint_name,
            config=EndpointCoreConfigInput(served_models=[served_model_input]),
            tags=tags,
        )
        if isinstance(endpoint, Mock):  # For testing
            served_model_output = ServedModelOutput(
                model_name=model_name,
                model_version=model_version,
                scale_to_zero_enabled=True,
                workload_size="Small",
            )
            endpoint = ServingEndpointDetailed(
                name=endpoint_name,
                pending_config=EndpointPendingConfig(served_models=[served_model_output]),
                tags=tags,
            )
        return endpoint

    def remove(endpoint: ServingEndpointDetailed) -> None:
        if endpoint.name:
            ws.serving_endpoints.delete(endpoint.name)

    yield from factory("Serving endpoint", create, remove)


@fixture
def make_feature_table(ws, make_random):
    def create():
        feature_table_name = make_random(6) + "." + make_random(6)
        table = ws.api_client.do(
            "POST",
            "/api/2.0/feature-store/feature-tables/create",
            body={"name": feature_table_name, "primary_keys": [{"name": "pk", "data_type": "string"}]},
        )
        return table['feature_table']

    def remove(table: dict):
        ws.api_client.do("DELETE", "/api/2.0/feature-store/feature-tables/delete", body={"name": table["name"]})

    yield from factory("Feature table", create, remove)



import io
import logging
import sys
from collections.abc import Callable, Generator
from pathlib import Path
from unittest.mock import Mock

from pytest import fixture
from databricks.labs.blueprint.paths import WorkspacePath
from databricks.sdk.service.workspace import ImportFormat, Language, RepoInfo
from databricks.sdk import WorkspaceClient

from databricks.labs.pytester.fixtures.baseline import factory


logger = logging.getLogger(__name__)
_DEFAULT_ENCODING = sys.getdefaultencoding()


@fixture
def make_notebook(ws, make_random, watchdog_purge_suffix) -> Generator[Callable[..., WorkspacePath], None, None]:
    """
    Returns a function to create Databricks Notebooks and clean them up after the test.
    The function returns [`os.PathLike` object](https://github.com/databrickslabs/blueprint?tab=readme-ov-file#python-native-pathlibpath-like-interfaces).

    Keyword arguments:
    * `path` (str, optional): The path of the notebook. Defaults to `dummy-*` notebook in current user's home folder.
    * `content` (str | bytes | io.BinaryIO, optional): The content of the notebook. Defaults to `print(1)` for Python and `SELECT 1` for SQL.
    * `language` (`databricks.sdk.service.workspace.Language`, optional): The language of the notebook. Defaults to `Language.PYTHON`.
    * `encoding` (`str`, optional): The file encoding. Defaults to `sys.getdefaultencoding()`.
    * `format` (`databricks.sdk.service.workspace.ImportFormat`, optional): The format of the notebook. Defaults to `ImportFormat.SOURCE`.
    * `overwrite` (bool, optional): Whether to overwrite the notebook if it already exists. Defaults to `False`.

    This example creates a notebook and verifies that `print(1)` is in the content:
    ```python
    def test_creates_some_notebook(make_notebook):
        notebook = make_notebook()
        assert "print(1)" in notebook.read_text()
    ```
    """

    def create(
        *,
        path: str | Path | None = None,
        content: str | bytes | io.BytesIO | None = None,
        language: Language | None = None,
        encoding: str | None = None,
        format: ImportFormat = ImportFormat.SOURCE,  # pylint:  disable=redefined-builtin
        overwrite: bool = False,
    ) -> WorkspacePath:
        encoding = encoding or _DEFAULT_ENCODING
        language = language or Language.PYTHON
        if language == Language.PYTHON:
            default_content = "print(1)"
        elif language == Language.SQL:
            default_content = "SELECT 1"
        else:
            raise ValueError(f"Unsupported language: {language}")
        current_user = ws.current_user.me()
        path = path or f"/Users/{current_user.user_name}/dummy-{make_random(4)}-{watchdog_purge_suffix}"
        workspace_path = WorkspacePath(ws, path)
        if '@' not in current_user.user_name:
            # If current user is a service principal added with `make_run_as`, there might be no home folder
            workspace_path.parent.mkdir(exist_ok=True)
        content = content or default_content
        if isinstance(content, str):
            content = io.BytesIO(content.encode(encoding))
        if isinstance(ws, Mock):  # For testing
            ws.workspace.download.return_value = content if isinstance(content, io.BytesIO) else io.BytesIO(content)
        ws.workspace.upload(path, content, language=language, format=format, overwrite=overwrite)
        logger.info(f"Created notebook: {workspace_path.as_uri()}")
        return workspace_path

    yield from factory("notebook", create, lambda path: path.unlink(missing_ok=True))


@fixture
def make_workspace_file(ws, make_random, watchdog_purge_suffix) -> Generator[Callable[..., WorkspacePath], None, None]:
    """
    Returns a function to create Databricks workspace file and clean up after the test.
    The function returns [`os.PathLike` object](https://github.com/databrickslabs/blueprint?tab=readme-ov-file#python-native-pathlibpath-like-interfaces).

    Keyword arguments:
    * `path` (str, optional): The path of the file. Defaults to `dummy-*` notebook in current user's home folder.
    * `content` (str | bytes, optional): The content of the file. Defaults to `print(1)` for Python and `SELECT 1` for SQL.
    * `language` (`databricks.sdk.service.workspace.Language`, optional): The language of the notebook. Defaults to `Language.PYTHON`.
    * `encoding` (`str`, optional): The file encoding. Defaults to `sys.getdefaultencoding()`.

    This example creates a notebook and verifies that the workspace path is an existing file with contents `print(1)`:
    ```python
    def test_create_file(make_workspace_file):
        workspace_file = make_workspace_file()
        assert workspace_file.is_file()
        assert "print(1)" in workspace_file.read_text()
    ```

    TODO:
    Merge functionality with `make_notebook` if `WorkspacePath` supports creating notebooks.
    """

    def create(
        *,
        path: str | Path | None = None,
        content: str | bytes | None = None,
        language: Language | None = None,
        encoding: str | None = None,
    ) -> WorkspacePath:
        language = language or Language.PYTHON
        if language == Language.PYTHON:
            default_content = "print(1)"
            suffix = ".py"
        elif language == Language.SQL:
            default_content = "SELECT 1"
            suffix = ".sql"
        else:
            raise ValueError(f"Unsupported language: {language}")
        current_user = ws.current_user.me()
        path = path or f"/Users/{current_user.user_name}/dummy-{make_random(4)}-{watchdog_purge_suffix}{suffix}"
        content = content or default_content
        encoding = encoding or _DEFAULT_ENCODING
        workspace_path = WorkspacePath(ws, path)
        if '@' not in current_user.user_name:
            # If current user is a service principal added with `make_run_as`, there might be no home folder
            workspace_path.parent.mkdir(exist_ok=True)
        if isinstance(content, bytes):
            workspace_path.write_bytes(content)
        else:
            workspace_path.write_text(content, encoding=encoding)
            content = content.encode(encoding)  # For testing
        if isinstance(ws, Mock):  # For testing
            ws.workspace.download.return_value = io.BytesIO(content)
        logger.info(f"Created file: {workspace_path.as_uri()}")
        return workspace_path

    yield from factory("file", create, lambda path: path.unlink(missing_ok=True))


@fixture
def make_directory(
    ws: WorkspaceClient, make_random, watchdog_purge_suffix
) -> Generator[Callable[..., WorkspacePath], None, None]:
    """
    Returns a function to create Databricks Workspace Folders and clean them up after the test.
    The function returns [`os.PathLike` object](https://github.com/databrickslabs/blueprint?tab=readme-ov-file#python-native-pathlibpath-like-interfaces).

    Keyword arguments:
    * `path` (str, optional): The path of the notebook. Defaults to `dummy-*` folder in current user's home folder.

    This example creates a folder and verifies that it contains a notebook:
    ```python
    def test_creates_some_folder_with_a_notebook(make_directory, make_notebook):
        folder = make_directory()
        notebook = make_notebook(path=folder / 'foo.py')
        files = [_.name for _ in folder.iterdir()]
        assert ['foo.py'] == files
        assert notebook.parent == folder
    ```
    """

    def create(*, path: str | Path | None = None) -> WorkspacePath:
        if path is None:
            path = f"~/dummy-{make_random(4)}-{watchdog_purge_suffix}"
        workspace_path = WorkspacePath(ws, path).expanduser()
        workspace_path.mkdir(exist_ok=True)
        logger.info(f"Created folder: {workspace_path.as_uri()}")
        return workspace_path

    yield from factory("directory", create, lambda ws_path: ws_path.rmdir(recursive=True))


@fixture
def make_repo(ws, make_random, watchdog_purge_suffix) -> Generator[Callable[..., RepoInfo], None, None]:
    """
    Returns a function to create Databricks Repos and clean them up after the test.
    The function returns a `databricks.sdk.service.workspace.RepoInfo` object.

    Keyword arguments:
    * `url` (str, optional): The URL of the repository.
    * `provider` (str, optional): The provider of the repository.
    * `path` (str, optional): The path of the repository. Defaults to `/Repos/{current_user}/sdk-{random}-{purge_suffix}`.

    Usage:
    ```python
    def test_repo(make_repo):
        logger.info(f"created {make_repo()}")
    ```
    """

    def create(*, url=None, provider=None, path=None, **kwargs) -> RepoInfo:
        if path is None:
            path = f"/Repos/{ws.current_user.me().user_name}/sdk-{make_random(4)}-{watchdog_purge_suffix}"
        if url is None:
            url = "https://github.com/shreyas-goenka/empty-repo.git"
        if provider is None:
            provider = "github"
        return ws.repos.create(url, provider, path=path, **kwargs)

    yield from factory("repo", create, lambda x: ws.repos.delete(x.id))



import pytest

from databricks.labs.pytester.fixtures.baseline import (
    ws,
    acc,
    make_random,
    product_info,
    log_workspace_link,
    log_account_link,
)
from databricks.labs.pytester.fixtures.sql import sql_backend, sql_exec, sql_fetch_all
from databricks.labs.pytester.fixtures.compute import (
    make_instance_pool,
    make_job,
    make_cluster,
    make_cluster_policy,
    make_pipeline,
    make_warehouse,
)
from databricks.labs.pytester.fixtures.iam import make_group, make_acc_group, make_user, make_run_as
from databricks.labs.pytester.fixtures.catalog import (
    make_udf,
    make_catalog,
    make_schema,
    make_table,
    make_storage_credential,
    make_volume,
)
from databricks.labs.pytester.fixtures.workspace import make_directory, make_workspace_file, make_notebook, make_repo
from databricks.labs.pytester.fixtures.permissions import (  # noqa
    make_cluster_policy_permissions,
    make_instance_pool_permissions,
    make_cluster_permissions,
    make_pipeline_permissions,
    make_job_permissions,
    make_notebook_permissions,
    make_directory_permissions,
    make_workspace_file_permissions,
    make_workspace_file_path_permissions,
    make_repo_permissions,
    make_authorization_permissions,
    make_warehouse_permissions,
    make_lakeview_dashboard_permissions,
    make_dashboard_permissions,
    make_alert_permissions,
    make_query_permissions,
    make_experiment_permissions,
    make_registered_model_permissions,
    make_serving_endpoint_permissions,
    make_feature_table_permissions,
)
from databricks.labs.pytester.fixtures.secrets import make_secret_scope, make_secret_scope_acl
from databricks.labs.pytester.fixtures.environment import debug_env, debug_env_name, env_or_skip, is_in_debug
from databricks.labs.pytester.fixtures.ml import make_experiment, make_model, make_serving_endpoint, make_feature_table
from databricks.labs.pytester.fixtures.redash import make_query
from databricks.labs.pytester.fixtures.watchdog import watchdog_remove_after, watchdog_purge_suffix
from databricks.labs.pytester.fixtures.connect import spark

__all__ = [
    'debug_env_name',
    'debug_env',
    'env_or_skip',
    'ws',
    'make_run_as',
    'acc',
    'spark',
    'sql_backend',
    'sql_exec',
    'sql_fetch_all',
    'make_random',
    'make_instance_pool',
    'make_instance_pool_permissions',
    'make_job',
    'make_job_permissions',
    'make_cluster',
    'make_cluster_permissions',
    'make_cluster_policy',
    'make_cluster_policy_permissions',
    'make_pipeline',
    'make_warehouse',
    'make_group',
    'make_acc_group',
    'make_user',
    'make_pipeline_permissions',
    'make_notebook',
    'make_notebook_permissions',
    'make_workspace_file',
    'make_directory',
    'make_directory_permissions',
    'make_repo',
    'make_repo_permissions',
    'make_workspace_file_permissions',
    'make_workspace_file_path_permissions',
    'make_secret_scope',
    'make_secret_scope_acl',
    'make_authorization_permissions',
    'make_udf',
    'make_catalog',
    'make_schema',
    'make_table',
    'make_storage_credential',
    'make_volume',
    'product_info',
    'make_model',
    'make_experiment',
    'make_experiment_permissions',
    'make_warehouse_permissions',
    'make_lakeview_dashboard_permissions',
    'log_workspace_link',
    'log_account_link',
    'make_dashboard_permissions',
    'make_alert_permissions',
    'make_query',
    'make_query_permissions',
    'make_registered_model_permissions',
    'make_serving_endpoint',
    'make_serving_endpoint_permissions',
    'make_feature_table',
    'make_feature_table_permissions',
    'watchdog_remove_after',
    'watchdog_purge_suffix',
    'is_in_debug',
]


def pytest_addoption(parser):
    group = parser.getgroup("helloworld")
    group.addoption(
        "--name",
        action="store",
        dest="name",
        default="World",
        help='Default "name" for hello().',
    )


@pytest.fixture
def hello(request):
    def _hello(name=None):
        if not name:
            name = request.config.getoption("name")
        return f"Hello {name}!"

    return _hello



import logging
import warnings
from collections.abc import Callable, Generator, Iterable
from datetime import timedelta

import pytest
from pytest import fixture
from databricks.sdk.credentials_provider import OAuthCredentialsProvider, OauthCredentialsStrategy
from databricks.sdk.oauth import ClientCredentials, Token
from databricks.sdk.service.oauth2 import CreateServicePrincipalSecretResponse
from databricks.labs.lsql import Row
from databricks.labs.lsql.backends import StatementExecutionBackend, SqlBackend
from databricks.sdk import AccountGroupsAPI, GroupsAPI, WorkspaceClient, AccountClient
from databricks.sdk.config import Config
from databricks.sdk.errors import ResourceConflict, NotFound
from databricks.sdk.retries import retried
from databricks.sdk.service import iam
from databricks.sdk.service.iam import (
    User,
    Group,
    ServicePrincipal,
    Patch,
    PatchOp,
    ComplexValue,
    PatchSchema,
    WorkspacePermission,
)

from databricks.labs.pytester.fixtures.baseline import factory

logger = logging.getLogger(__name__)


@fixture
def make_user(ws, make_random, log_workspace_link, watchdog_purge_suffix):
    """
    This fixture returns a function that creates a Databricks workspace user
    and removes it after the test is complete. In case of random naming conflicts,
    the fixture will retry the creation process for 30 seconds. Returns an instance
    of `databricks.sdk.service.iam.User`. Usage:

    ```python
    def test_new_user(make_user, ws):
        new_user = make_user()
        home_dir = ws.workspace.get_status(f"/Users/{new_user.user_name}")
        assert home_dir.object_type == ObjectType.DIRECTORY
    ```
    """

    @retried(on=[ResourceConflict], timeout=timedelta(seconds=30))
    def create(**kwargs) -> User:
        user_name = f"dummy-{make_random(4)}-{watchdog_purge_suffix}@example.com".lower()
        user = ws.users.create(user_name=user_name, **kwargs)
        log_workspace_link(user.user_name, f'settings/workspace/identity-and-access/users/{user.id}')
        return user

    yield from factory("workspace user", create, lambda item: ws.users.delete(item.id))


@fixture
def make_group(ws: WorkspaceClient, make_random, watchdog_purge_suffix):
    """
    This fixture provides a function to manage Databricks workspace groups. Groups can be created with specified
    members and roles, and they will be deleted after the test is complete. Deals with eventual consistency issues by
    retrying the creation process for 30 seconds and then waiting for up to 3 minutes for the group to be provisioned.
    Returns an instance of `databricks.sdk.service.iam.Group`.

    Keyword arguments:
    * `members` (list of strings): A list of user IDs to add to the group.
    * `roles` (list of strings): A list of roles to assign to the group.
    * `display_name` (str): The display name of the group.
    * `entitlements` (list of strings): A list of entitlements to assign to the group.

    The following example creates a group with a single member and independently verifies that the group was created:

    ```python
    def test_new_group(make_group, make_user, ws):
        user = make_user()
        group = make_group(members=[user.id])
        loaded = ws.groups.get(group.id)
        assert group.display_name == loaded.display_name
        assert group.members == loaded.members
    ```
    """
    yield from _make_group("workspace group", ws.config, ws.groups, make_random, watchdog_purge_suffix)


@fixture
def make_acc_group(acc, make_random, watchdog_purge_suffix):
    """
    This fixture provides a function to manage Databricks account groups. Groups can be created with
    specified members and roles, and they will be deleted after the test is complete.

    Has the same arguments and behavior as [`make_group` fixture](#make_group-fixture) but uses the account
    client instead of the workspace client.

    Example usage:
    ```python
    def test_new_account_group(make_acc_group, acc):
        group = make_acc_group()
        loaded = acc.groups.get(group.id)
        assert group.display_name == loaded.display_name
    ```
    """
    yield from _make_group("account group", acc.config, acc.groups, make_random, watchdog_purge_suffix)


def _scim_values(ids: list[str]) -> list[iam.ComplexValue]:
    return [iam.ComplexValue(value=x) for x in ids]


def _wait_group_provisioned(interface: AccountGroupsAPI | GroupsAPI, group: Group) -> None:
    """Wait for a group to be visible via the supplied group interface.

    Due to consistency issues in the group-management APIs, new groups are not always visible in a consistent manner
    after being created or modified. This method can be used to mitigate against this by checking that a group:

     - Is visible via the `.get()` interface;
     - Is visible via the `.list()` interface that enumerates groups.

    Visibility is assumed when 2 calls in a row return the expected results.

    Args:
          interface: the group-management interface to use for checking whether the groups are visible.
          group: the group whose visibility should be verified.
    Raises:
          NotFound: this is thrown if it takes longer than 90 seconds for the group to become visible via the
          management interface.
    """
    # Use double-checking to try and compensate for the lack of monotonic consistency with the group-management
    # interfaces: two subsequent calls need to succeed for us to proceed. (This is probabilistic, and not a guarantee.)
    # The REST API internals cache things for up to 60s, and we see times close to this during tests. The retry timeout
    # reflects this: if it's taking much longer then something else is wrong.
    group_id = group.id
    assert group_id is not None

    @retried(on=[NotFound], timeout=timedelta(seconds=90))
    def _double_get_group() -> None:
        interface.get(group_id)
        interface.get(group_id)

    def _check_group_in_listing() -> None:
        found_groups = interface.list(attributes="id", filter=f'id eq "{group_id}"')
        found_ids = {found_group.id for found_group in found_groups}
        if group_id not in found_ids:
            msg = f"Group id not (yet) found in group listing: {group_id}"
            raise NotFound(msg)

    @retried(on=[NotFound], timeout=timedelta(seconds=90))
    def _double_check_group_in_listing() -> None:
        _check_group_in_listing()
        _check_group_in_listing()

    _double_get_group()
    _double_check_group_in_listing()


def _make_group(
    name: str, cfg: Config, interface, make_random, watchdog_purge_suffix
) -> Generator[Callable[..., Group], None, None]:
    _not_specified = object()

    @retried(on=[ResourceConflict], timeout=timedelta(seconds=30))
    def create(
        *,
        members: list[str] | None = None,
        roles: list[str] | None = None,
        entitlements: list[str] | None = None,
        display_name: str | None = None,
        wait_for_provisioning: bool | object = _not_specified,
        **kwargs,
    ):
        kwargs["display_name"] = (
            f"sdk-{make_random(4)}-{watchdog_purge_suffix}" if display_name is None else display_name
        )
        if members is not None:
            kwargs["members"] = _scim_values(members)
        if roles is not None:
            kwargs["roles"] = _scim_values(roles)
        if entitlements is not None:
            kwargs["entitlements"] = _scim_values(entitlements)
        if wait_for_provisioning is not _not_specified:
            warnings.warn(
                "Specifying wait_for_provisioning when making a group is deprecated; we always wait.",
                DeprecationWarning,
                # Call stack is: create()[iam.py], wrapper()[retries.py], inner()[baseline.py], client_code
                stacklevel=4,
            )
        # TODO: REQUEST_LIMIT_EXCEEDED: GetUserPermissionsRequest RPC token bucket limit has been exceeded.
        group = interface.create(**kwargs)
        if cfg.is_account_client:
            logger.info(f"Account group {group.display_name}: {cfg.host}/users/groups/{group.id}/members")
        else:
            logger.info(f"Workspace group {group.display_name}: {cfg.host}#setting/accounts/groups/{group.id}")

        _wait_group_provisioned(interface, group)

        return group

    yield from factory(name, create, lambda item: interface.delete(item.id))


class RunAs:
    def __init__(self, service_principal: ServicePrincipal, workspace_client: WorkspaceClient, env_or_skip):
        self._service_principal = service_principal
        self._workspace_client = workspace_client
        self._env_or_skip = env_or_skip

    @property
    def ws(self):
        return self._workspace_client

    @property
    def sql_backend(self) -> SqlBackend:
        # TODO: Switch to `__getattr__` + `SubRequest` to get a generic way of initializing all workspace fixtures.
        # This will allow us to remove the `sql_backend` fixture and make the `RunAs` class more generic.
        # It turns out to be more complicated than it first appears, because we don't get these at pytest.collect phase.
        warehouse_id = self._env_or_skip("DATABRICKS_WAREHOUSE_ID")
        return StatementExecutionBackend(self._workspace_client, warehouse_id)

    def sql_exec(self, statement: str) -> None:
        return self.sql_backend.execute(statement)

    def sql_fetch_all(self, statement: str) -> Iterable[Row]:
        return self.sql_backend.fetch(statement)

    def __getattr__(self, item: str):
        if item in self.__dict__:
            return self.__dict__[item]
        fixture_value = self._request.getfixturevalue(item)
        return fixture_value

    @property
    def display_name(self) -> str:
        assert self._service_principal.display_name is not None
        return self._service_principal.display_name

    @property
    def application_id(self) -> str:
        assert self._service_principal.application_id is not None
        return self._service_principal.application_id

    @property
    def id(self) -> str:
        assert self._service_principal.id is not None
        return self._service_principal.id

    def __repr__(self):
        return f'RunAs({self.display_name})'


def _make_workspace_client(
    ws: WorkspaceClient,
    created_secret: CreateServicePrincipalSecretResponse,
    service_principal: ServicePrincipal,
) -> WorkspaceClient:
    oidc = ws.config.oidc_endpoints
    assert oidc is not None, 'OIDC is required'
    application_id = service_principal.application_id
    secret_value = created_secret.secret
    assert application_id is not None
    assert secret_value is not None

    token_source = ClientCredentials(
        client_id=application_id,
        client_secret=secret_value,
        token_url=oidc.token_endpoint,
        scopes=["all-apis"],
        use_header=True,
    )

    def inner() -> dict[str, str]:
        inner_token = token_source.token()
        return {'Authorization': f'{inner_token.token_type} {inner_token.access_token}'}

    def token() -> Token:
        return token_source.token()

    credentials_provider = OAuthCredentialsProvider(inner, token)
    credentials_strategy = OauthCredentialsStrategy('oauth-m2m', lambda _: credentials_provider)
    ws_as_spn = WorkspaceClient(host=ws.config.host, credentials_strategy=credentials_strategy)
    return ws_as_spn


@fixture
def make_run_as(acc: AccountClient, ws: WorkspaceClient, make_random, env_or_skip, log_account_link, is_in_debug):
    """
    This fixture provides a function to create an account service principal via [`acc` fixture](#acc-fixture) and
    assign it to a workspace. The service principal is removed after the test is complete. The service principal is
    created with a random display name and assigned to the workspace with the default permissions.

    Use the `account_groups` argument to assign the service principal to account groups, which have the required
    permissions to perform a specific action.

    Example:

    ```python
    def test_run_as_lower_privilege_user(make_run_as, ws):
        run_as = make_run_as(account_groups=['account.group.name'])
        through_query = next(run_as.sql_fetch_all("SELECT CURRENT_USER() AS my_name"))
        me = ws.current_user.me()
        assert me.user_name != through_query.my_name
    ```

    Returned object has the following properties:
    * `ws`: Workspace client that is authenticated as the ephemeral service principal.
    * `sql_backend`: SQL backend that is authenticated as the ephemeral service principal.
    * `sql_exec`: Function to execute a SQL statement on behalf of the ephemeral service principal.
    * `sql_fetch_all`: Function to fetch all rows from a SQL statement on behalf of the ephemeral service principal.
    * `display_name`: Display name of the ephemeral service principal.
    * `application_id`: Application ID of the ephemeral service principal.
    * if you want to have other fixtures available in the context of the ephemeral service principal, you can override
      the [`ws` fixture](#ws-fixture) on the file level, which would make all workspace fixtures provided by this
      plugin to run as lower privilege ephemeral service principal. You cannot combine it with the account-admin-level
      principal you're using to create the ephemeral principal.

    Example:

    ```python
    from pytest import fixture

    @fixture
    def ws(make_run_as):
        run_as = make_run_as(account_groups=['account.group.used.for.all.tests.in.this.file'])
        return run_as.ws

    def test_creating_notebook_on_behalf_of_ephemeral_principal(make_notebook):
        notebook = make_notebook()
        assert notebook.exists()
    ```

    This fixture currently doesn't work with Databricks Metadata Service authentication on Azure Databricks.
    """

    if ws.config.auth_type == 'metadata-service' and ws.config.is_azure:
        # TODO: fix `invalid_scope: AADSTS1002012: The provided value for scope all-apis is not valid.` error
        #
        # We're having issues with the Azure Metadata Service and service principals. The error message is:
        # Client credential flows must have a scope value with /.default suffixed to the resource identifier
        # (application ID URI)
        pytest.skip('Azure Metadata Service does not support service principals')

    def create(*, account_groups: list[str] | None = None):
        workspace_id = ws.get_workspace_id()
        service_principal = acc.service_principals.create(display_name=f'spn-{make_random()}')
        assert service_principal.id is not None
        created_secret = acc.service_principal_secrets.create(service_principal.id)
        if account_groups:
            group_mapping = {}
            for group in acc.groups.list(attributes='id,displayName'):
                if group.id is None:
                    continue
                group_mapping[group.display_name] = group.id
            for group_name in account_groups:
                if group_name not in group_mapping:
                    raise ValueError(f'Group {group_name} does not exist')
                group_id = group_mapping[group_name]
                acc.groups.patch(
                    group_id,
                    operations=[
                        Patch(PatchOp.ADD, 'members', [ComplexValue(value=str(service_principal.id)).as_dict()]),
                    ],
                    schemas=[PatchSchema.URN_IETF_PARAMS_SCIM_API_MESSAGES_2_0_PATCH_OP],
                )
        permissions = [WorkspacePermission.USER]
        acc.workspace_assignment.update(workspace_id, int(service_principal.id), permissions=permissions)
        ws_as_spn = _make_workspace_client(ws, created_secret, service_principal)

        log_account_link('account service principal', f'users/serviceprincipals/{service_principal.id}')

        return RunAs(service_principal, ws_as_spn, env_or_skip)

    def remove(run_as: RunAs):
        service_principal_id = run_as._service_principal.id  # pylint: disable=protected-access
        assert service_principal_id is not None
        acc.service_principals.delete(service_principal_id)

    yield from factory("service principal", create, remove)



import pytest
from databricks.sdk.errors import InvalidParameterValue
from databricks.sdk.service import iam
from databricks.sdk.service.iam import PermissionLevel
from databricks.sdk.service.sql import GetResponse, ObjectTypePlural, AccessControl
from databricks.sdk.service.sql import PermissionLevel as RedashPermissionLevel

from databricks.labs.pytester.fixtures.baseline import factory

# pylint: disable=too-complex


class _PermissionsChange:
    def __init__(self, object_id: str, before: list[iam.AccessControlRequest], after: list[iam.AccessControlRequest]):
        self.object_id = object_id
        self.before = before
        self.after = after

    @staticmethod
    def _principal(acr: iam.AccessControlRequest) -> str:
        if acr.user_name is not None:
            return f"user_name {acr.user_name}"
        if acr.group_name is not None:
            return f"group_name {acr.group_name}"
        return f"service_principal_name {acr.service_principal_name}"

    def _list(self, acl: list[iam.AccessControlRequest]):
        return ", ".join(f"{self._principal(_)} {_.permission_level.value}" for _ in acl)

    def __repr__(self):
        return f"{self.object_id} [{self._list(self.before)}] -> [{self._list(self.after)}]"


class _RedashPermissionsChange:
    def __init__(self, object_id: str, before: list[AccessControl], after: list[AccessControl]):
        self.object_id = object_id
        self.before = before
        self.after = after

    @staticmethod
    def _principal(acr: AccessControl) -> str:
        if acr.user_name is not None:
            return f"user_name {acr.user_name}"
        return f"group_name {acr.group_name}"

    def _list(self, acl: list[AccessControl]):
        return ", ".join(f"{self._principal(_)} {_.permission_level.value}" for _ in acl)

    def __repr__(self):
        return f"{self.object_id} [{self._list(self.before)}] -> [{self._list(self.after)}]"


def _make_permissions_factory(name, resource_type, levels, id_retriever):
    def _non_inherited(acl: iam.ObjectPermissions):
        out: list[iam.AccessControlRequest] = []
        assert acl.access_control_list is not None
        for access_control in acl.access_control_list:
            if not access_control.all_permissions:
                continue
            for permission in access_control.all_permissions:
                if not permission.inherited:
                    continue
                out.append(
                    iam.AccessControlRequest(
                        permission_level=permission.permission_level,
                        group_name=access_control.group_name,
                        user_name=access_control.user_name,
                        service_principal_name=access_control.service_principal_name,
                    )
                )
        return out

    def _make_permissions(ws):
        def create(
            *,
            object_id: str,
            permission_level: iam.PermissionLevel | None = None,
            group_name: str | None = None,
            user_name: str | None = None,
            service_principal_name: str | None = None,
            access_control_list: list[iam.AccessControlRequest] | None = None,
        ):
            nothing_specified = permission_level is None and access_control_list is None
            both_specified = permission_level is not None and access_control_list is not None
            if nothing_specified or both_specified:
                msg = "either permission_level or access_control_list has to be specified"
                raise ValueError(msg)

            object_id = id_retriever(ws, object_id)
            initial = _non_inherited(ws.permissions.get(resource_type, object_id))
            if access_control_list is None:
                if permission_level not in levels:
                    assert permission_level is not None
                    names = ", ".join(_.value for _ in levels)
                    msg = f"invalid permission level: {permission_level.value}. Valid levels: {names}"
                    raise ValueError(msg)

                access_control_list = []
                if group_name is not None:
                    access_control_list.append(
                        iam.AccessControlRequest(
                            group_name=group_name,
                            permission_level=permission_level,
                        )
                    )
                if user_name is not None:
                    access_control_list.append(
                        iam.AccessControlRequest(
                            user_name=user_name,
                            permission_level=permission_level,
                        )
                    )
                if service_principal_name is not None:
                    access_control_list.append(
                        iam.AccessControlRequest(
                            service_principal_name=service_principal_name,
                            permission_level=permission_level,
                        )
                    )
            ws.permissions.update(resource_type, object_id, access_control_list=access_control_list)
            return _PermissionsChange(object_id, initial, access_control_list)

        def remove(change: _PermissionsChange):
            try:
                ws.permissions.set(resource_type, change.object_id, access_control_list=change.before)
            except InvalidParameterValue:
                pass

        yield from factory(f"{name} permissions", create, remove)

    return _make_permissions


def _make_redash_permissions_factory(name, resource_type, levels, id_retriever):
    def _non_inherited(acl: GetResponse):
        out: list[AccessControl] = []
        assert acl.access_control_list is not None
        for access_control in acl.access_control_list:
            out.append(
                AccessControl(
                    permission_level=access_control.permission_level,
                    group_name=access_control.group_name,
                    user_name=access_control.user_name,
                )
            )
        return out

    def _make_permissions(ws):
        def create(
            *,
            object_id: str,
            permission_level: RedashPermissionLevel | None = None,
            group_name: str | None = None,
            user_name: str | None = None,
            access_control_list: list[AccessControl] | None = None,
        ):
            nothing_specified = permission_level is None and access_control_list is None
            both_specified = permission_level is not None and access_control_list is not None
            if nothing_specified or both_specified:
                msg = "either permission_level or access_control_list has to be specified"
                raise ValueError(msg)

            object_id = id_retriever(ws, object_id)
            initial = _non_inherited(ws.dbsql_permissions.get(resource_type, object_id))

            if access_control_list is None:
                if permission_level not in levels:
                    assert permission_level is not None
                    names = ", ".join(_.value for _ in levels)
                    msg = f"invalid permission level: {permission_level.value}. Valid levels: {names}"
                    raise ValueError(msg)

                access_control_list = []
                if group_name is not None:
                    access_control_list.append(
                        AccessControl(
                            group_name=group_name,
                            permission_level=permission_level,
                        )
                    )
                if user_name is not None:
                    access_control_list.append(
                        AccessControl(
                            user_name=user_name,
                            permission_level=permission_level,
                        )
                    )

            ws.dbsql_permissions.set(resource_type, object_id, access_control_list=access_control_list)
            return _RedashPermissionsChange(object_id, initial, access_control_list)

        def remove(change: _RedashPermissionsChange):
            ws.dbsql_permissions.set(
                ObjectTypePlural(resource_type),
                change.object_id,
                access_control_list=change.before,
            )

        yield from factory(f"{name} permissions", create, remove)

    return _make_permissions


def _simple(_, object_id):
    return object_id


def _path(ws, path):
    return ws.workspace.get_status(path).object_id


make_cluster_policy_permissions = pytest.fixture(
    _make_permissions_factory(
        "cluster_policy",
        "cluster-policies",
        [
            PermissionLevel.CAN_USE,
        ],
        _simple,
    )
)

make_instance_pool_permissions = pytest.fixture(
    _make_permissions_factory(
        "instance_pool",
        "instance-pools",
        [
            PermissionLevel.CAN_ATTACH_TO,
            PermissionLevel.CAN_MANAGE,
        ],
        _simple,
    )
)
make_cluster_permissions = pytest.fixture(
    _make_permissions_factory(
        "cluster",
        "clusters",
        [
            PermissionLevel.CAN_ATTACH_TO,
            PermissionLevel.CAN_RESTART,
            PermissionLevel.CAN_MANAGE,
        ],
        _simple,
    )
)
make_pipeline_permissions = pytest.fixture(
    _make_permissions_factory(
        "pipeline",
        "pipelines",
        [
            PermissionLevel.CAN_VIEW,
            PermissionLevel.CAN_RUN,
            PermissionLevel.CAN_MANAGE,
            PermissionLevel.IS_OWNER,  # cannot be a group
        ],
        _simple,
    )
)
make_job_permissions = pytest.fixture(
    _make_permissions_factory(
        "job",
        "jobs",
        [
            PermissionLevel.CAN_VIEW,
            PermissionLevel.CAN_MANAGE_RUN,
            PermissionLevel.CAN_MANAGE,
            PermissionLevel.IS_OWNER,  # cannot be a group
        ],
        _simple,
    )
)
make_notebook_permissions = pytest.fixture(
    _make_permissions_factory(
        "notebook",
        "notebooks",
        [
            PermissionLevel.CAN_READ,
            PermissionLevel.CAN_RUN,
            PermissionLevel.CAN_EDIT,
            PermissionLevel.CAN_MANAGE,
        ],
        _path,
    )
)
make_directory_permissions = pytest.fixture(
    _make_permissions_factory(
        "directory",
        "directories",
        [
            PermissionLevel.CAN_READ,
            PermissionLevel.CAN_RUN,
            PermissionLevel.CAN_EDIT,
            PermissionLevel.CAN_MANAGE,
        ],
        _path,
    )
)
make_workspace_file_permissions = pytest.fixture(
    _make_permissions_factory(
        "workspace_file",
        "files",
        [
            PermissionLevel.CAN_READ,
            PermissionLevel.CAN_RUN,
            PermissionLevel.CAN_EDIT,
            PermissionLevel.CAN_MANAGE,
        ],
        _simple,
    )
)
make_workspace_file_path_permissions = pytest.fixture(
    _make_permissions_factory(
        "workspace_file_path",
        "files",
        [
            PermissionLevel.CAN_READ,
            PermissionLevel.CAN_RUN,
            PermissionLevel.CAN_EDIT,
            PermissionLevel.CAN_MANAGE,
        ],
        _path,
    )
)
make_repo_permissions = pytest.fixture(
    _make_permissions_factory(
        "repo",
        "repos",
        [
            PermissionLevel.CAN_READ,
            PermissionLevel.CAN_RUN,
            PermissionLevel.CAN_EDIT,
            PermissionLevel.CAN_MANAGE,
        ],
        _path,
    )
)
make_authorization_permissions = pytest.fixture(
    _make_permissions_factory(
        "authorization",
        "authorization",
        [
            PermissionLevel.CAN_USE,
        ],
        _simple,
    )
)
make_warehouse_permissions = pytest.fixture(
    _make_permissions_factory(
        "warehouse",
        "sql/warehouses",
        [
            PermissionLevel.CAN_USE,
            PermissionLevel.CAN_MANAGE,
        ],
        _simple,
    )
)
make_lakeview_dashboard_permissions = pytest.fixture(
    _make_permissions_factory(
        "lakeview_dashboard",
        "dashboards",
        # The `CAN_READ` permission is consistent with the documentation (see below),
        # but not with the databricks UI as it shows `CAN_VIEW` instead.
        # https://docs.databricks.com/en/dashboards/tutorials/manage-permissions.html#get-workspace-object-permission-levels
        [
            PermissionLevel.CAN_EDIT,
            PermissionLevel.CAN_RUN,
            PermissionLevel.CAN_MANAGE,
            PermissionLevel.CAN_READ,
        ],
        _simple,
    )
)
make_dashboard_permissions = pytest.fixture(
    _make_redash_permissions_factory(
        "dashboard",
        ObjectTypePlural.DASHBOARDS,
        [
            RedashPermissionLevel.CAN_EDIT,
            RedashPermissionLevel.CAN_RUN,
            RedashPermissionLevel.CAN_MANAGE,
            RedashPermissionLevel.CAN_VIEW,
        ],
        _simple,
    )
)
make_alert_permissions = pytest.fixture(
    _make_redash_permissions_factory(
        "alert",
        ObjectTypePlural.ALERTS,
        [
            RedashPermissionLevel.CAN_EDIT,
            RedashPermissionLevel.CAN_RUN,
            RedashPermissionLevel.CAN_MANAGE,
            RedashPermissionLevel.CAN_VIEW,
        ],
        _simple,
    )
)
make_query_permissions = pytest.fixture(
    _make_redash_permissions_factory(
        "query",
        ObjectTypePlural.QUERIES,
        [
            RedashPermissionLevel.CAN_EDIT,
            RedashPermissionLevel.CAN_RUN,
            RedashPermissionLevel.CAN_MANAGE,
            RedashPermissionLevel.CAN_VIEW,
        ],
        _simple,
    )
)
make_experiment_permissions = pytest.fixture(
    _make_permissions_factory(
        "experiment",
        "experiments",
        [
            PermissionLevel.CAN_READ,
            PermissionLevel.CAN_EDIT,
            PermissionLevel.CAN_MANAGE,
        ],
        _simple,
    )
)
make_registered_model_permissions = pytest.fixture(
    _make_permissions_factory(
        "registered_model",
        "registered-models",
        [
            PermissionLevel.CAN_READ,
            PermissionLevel.CAN_EDIT,
            PermissionLevel.CAN_MANAGE_STAGING_VERSIONS,
            PermissionLevel.CAN_MANAGE_PRODUCTION_VERSIONS,
            PermissionLevel.CAN_MANAGE,
        ],
        _simple,
    )
)
make_serving_endpoint_permissions = pytest.fixture(
    _make_permissions_factory(
        "serving_endpoint",
        "serving-endpoints",
        [
            PermissionLevel.CAN_VIEW,
            PermissionLevel.CAN_MANAGE,
            PermissionLevel.CAN_QUERY,
        ],
        _simple,
    )
)
make_feature_table_permissions = pytest.fixture(
    _make_permissions_factory(
        "feature_table",
        "feature-tables",
        [
            PermissionLevel.CAN_VIEW_METADATA,
            PermissionLevel.CAN_EDIT_METADATA,
            PermissionLevel.CAN_MANAGE,
        ],
        _simple,
    )
)



import json
import logging
from collections.abc import Generator, Callable
from unittest.mock import Mock

from pytest import fixture
from databricks.labs.blueprint.commands import CommandExecutor
from databricks.sdk.errors import DatabricksError
from databricks.sdk.service.catalog import (
    FunctionInfo,
    SchemaInfo,
    TableInfo,
    TableType,
    DataSourceFormat,
    CatalogInfo,
    StorageCredentialInfo,
    AwsIamRoleRequest,
    AzureServicePrincipal,
    VolumeInfo,
    VolumeType,
)
from databricks.sdk.service.compute import Language
from databricks.labs.pytester.fixtures.baseline import factory

logger = logging.getLogger(__name__)


# TODO: replace with LSQL implementation
def escape_sql_identifier(path: str, *, maxsplit: int = 2) -> str:
    """
    Escapes the path components to make them SQL safe.

    Args:
        path (str): The dot-separated path of a catalog object.
        maxsplit (int): The maximum number of splits to perform.

    Returns:
         str: The path with all parts escaped in backticks.
    """
    if not path:
        return path
    parts = path.split(".", maxsplit=maxsplit)
    escaped = [f"`{part.strip('`').replace('`', '``')}`" for part in parts]
    return ".".join(escaped)


@fixture
# pylint: disable-next=too-many-statements,too-complex
def make_table(
    sql_backend,
    make_schema,
    make_random,
    log_workspace_link,
    watchdog_remove_after,
) -> Generator[Callable[..., TableInfo], None, None]:
    """
    Create a table and return its info. Remove it after the test. Returns instance of `databricks.sdk.service.catalog.TableInfo`.

    Keyword Arguments:
    * `catalog_name` (str): The name of the catalog where the table will be created. Default is `hive_metastore`.
    * `schema_name` (str): The name of the schema where the table will be created. Default is a random string.
    * `name` (str): The name of the table. Default is a random string.
    * `ctas` (str): The CTAS statement to create the table. Default is `None`.
    * `non_delta` (bool): If `True`, the table will be created as a non-delta table. Default is `False`.
    * `external` (bool): If `True`, the table will be created as an external table. Default is `False`.
    * `external_csv` (str): The location of the external CSV table. Default is `None`.
    * `external_delta` (str): The location of the external Delta table. Default is `None`.
    * `view` (bool): If `True`, the table will be created as a view. Default is `False`.
    * `tbl_properties` (dict): The table properties. Default is `None`.
    * `hiveserde_ddl` (str): The DDL statement to create the table. Default is `None`.
    * `storage_override` (str): The storage location override. Default is `None`.
    * `columns` (list): The list of columns. Default is `None`.

    Usage:
    ```python
    def test_catalog_fixture(make_catalog, make_schema, make_table):
        from_catalog = make_catalog()
        from_schema = make_schema(catalog_name=from_catalog.name)
        from_table_1 = make_table(catalog_name=from_catalog.name, schema_name=from_schema.name)
        logger.info(f"Created new schema: {from_table_1}")
    ```
    """

    def generate_sql_schema(columns: list[tuple[str, str]]) -> str:
        """Generate a SQL schema from columns."""
        schema = "("
        for index, (col_name, type_name) in enumerate(columns):
            schema += escape_sql_identifier(col_name or str(index), maxsplit=0)
            schema += f" {type_name}, "
        schema = schema[:-2] + ")"  # Remove the last ', '
        return schema

    def generate_sql_column_casting(existing_columns: list[tuple[str, str]], new_columns: list[tuple[str, str]]) -> str:
        """Generate the SQL to cast columns"""
        if len(new_columns) > len(existing_columns):
            raise ValueError(f"Too many columns: {new_columns}")
        select_expressions = []
        for index, ((existing_name, _), (new_name, new_type)) in enumerate(zip(existing_columns, new_columns)):
            column_name_new = escape_sql_identifier(new_name or str(index), maxsplit=0)
            select_expression = f"CAST({existing_name} AS {new_type}) AS {column_name_new}"
            select_expressions.append(select_expression)
        select = ", ".join(select_expressions)
        return select

    def create(  # pylint: disable=too-many-locals,too-many-arguments,too-many-statements
        *,
        catalog_name="hive_metastore",
        schema_name: str | None = None,
        name: str | None = None,
        ctas: str | None = None,
        non_delta: bool = False,
        external: bool = False,
        external_csv: str | None = None,
        external_delta: str | None = None,
        view: bool = False,
        tbl_properties: dict[str, str] | None = None,
        hiveserde_ddl: str | None = None,
        storage_override: str | None = None,
        columns: list[tuple[str, str]] | None = None,
    ) -> TableInfo:
        if schema_name is None:
            schema = make_schema(catalog_name=catalog_name)
            catalog_name = schema.catalog_name
            schema_name = schema.name
        name = name or f"dummy_t{make_random(4).lower()}"
        table_type: TableType | None = None
        data_source_format = None
        storage_location = None
        view_text = None
        full_name = f"{catalog_name}.{schema_name}.{name}".lower()
        ddl = f'CREATE {"VIEW" if view else "TABLE"} {full_name}'
        if columns is None:
            schema = "(id INT, value STRING)"
        else:
            schema = generate_sql_schema(columns)
        if view:
            table_type = TableType.VIEW
            view_text = ctas
        if ctas is not None:
            # temporary (if not view)
            ddl = f"{ddl} AS {ctas}"
        elif non_delta:
            table_type = TableType.EXTERNAL  # pylint: disable=redefined-variable-type
            data_source_format = DataSourceFormat.JSON
            # DBFS locations are not purged; no suffix necessary.
            storage_location = f"dbfs:/tmp/{name}"
            if columns is None:
                select = "*"
            else:
                # These are the columns from the JSON dataset below
                dataset_columns = [
                    ('calories_burnt', 'STRING'),
                    ('device_id', 'STRING'),
                    ('id', 'STRING'),
                    ('miles_walked', 'STRING'),
                    ('num_steps', 'STRING'),
                    ('timestamp', 'STRING'),
                    ('user_id', 'STRING'),
                    ('value', 'STRING'),
                ]
                select = generate_sql_column_casting(dataset_columns, columns)
            # Modified, otherwise it will identify the table as a DB Dataset
            ddl = (
                f"{ddl} USING json location '{storage_location}' as SELECT {select} FROM "
                f"JSON.`dbfs:/databricks-datasets/iot-stream/data-device`"
            )
        elif external_csv is not None:
            table_type = TableType.EXTERNAL
            data_source_format = DataSourceFormat.CSV
            storage_location = external_csv
            ddl = f"{ddl} {schema} USING CSV OPTIONS (header=true) LOCATION '{storage_location}'"
        elif external_delta is not None:
            table_type = TableType.EXTERNAL
            data_source_format = DataSourceFormat.DELTA
            storage_location = external_delta
            ddl = f"{ddl} {schema} LOCATION '{storage_location}'"
        elif external:
            # external table
            table_type = TableType.EXTERNAL
            data_source_format = DataSourceFormat.DELTASHARING
            url = "s3a://databricks-datasets-oregon/delta-sharing/share/open-datasets.share"
            storage_location = f"{url}#delta_sharing.default.lending_club"
            ddl = f"{ddl} USING deltaSharing LOCATION '{storage_location}'"
        else:
            # managed table
            table_type = TableType.MANAGED
            data_source_format = DataSourceFormat.DELTA
            storage_location = f"dbfs:/user/hive/warehouse/{schema_name}/{name}"
            ddl = f"{ddl} {schema}"
        if tbl_properties:
            tbl_properties.update({"RemoveAfter": watchdog_remove_after})
        else:
            tbl_properties = {"RemoveAfter": watchdog_remove_after}

        str_properties = ",".join([f" '{k}' = '{v}' " for k, v in tbl_properties.items()])

        # table properties fails with CTAS statements
        alter_table_tbl_properties = ""
        if ctas or non_delta:
            alter_table_tbl_properties = (
                f'ALTER {"VIEW" if view else "TABLE"} {full_name} SET TBLPROPERTIES ({str_properties})'
            )
        else:
            ddl = f"{ddl} TBLPROPERTIES ({str_properties})"

        if hiveserde_ddl:
            ddl = hiveserde_ddl
            data_source_format = None
            table_type = TableType.EXTERNAL
            storage_location = storage_override

        sql_backend.execute(ddl)

        # CTAS AND NON_DELTA does not support TBLPROPERTIES
        if ctas or non_delta:
            sql_backend.execute(alter_table_tbl_properties)

        table_info = TableInfo(
            catalog_name=catalog_name,
            schema_name=schema_name,
            name=name,
            full_name=full_name,
            properties=tbl_properties,
            storage_location=storage_location,
            table_type=table_type,
            view_definition=view_text,
            data_source_format=data_source_format,
        )
        path = f'explore/data/{table_info.catalog_name}/{table_info.schema_name}/{table_info.name}'
        log_workspace_link(f'{table_info.full_name} schema', path)
        return table_info

    def remove(table_info: TableInfo):
        try:
            sql_backend.execute(f"DROP TABLE IF EXISTS {table_info.full_name}")
        except DatabricksError as e:
            if "Cannot drop a view" in str(e):
                sql_backend.execute(f"DROP VIEW IF EXISTS {table_info.full_name}")
            else:
                raise e

    yield from factory("table", create, remove)


@fixture
def make_schema(
    sql_backend, make_random, log_workspace_link, watchdog_remove_after
) -> Generator[Callable[..., SchemaInfo], None, None]:
    """
    Create a schema and return its info. Remove it after the test. Returns instance of `databricks.sdk.service.catalog.SchemaInfo`.

    Keyword Arguments:
    * `catalog_name` (str): The name of the catalog where the schema will be created. Default is `hive_metastore`.
    * `name` (str): The name of the schema. Default is a random string.
    * `location` (str): The path to the location if it should be a managed schema.

    Usage:
    ```python
    def test_catalog_fixture(make_catalog, make_schema, make_table):
        from_catalog = make_catalog()
        from_schema = make_schema(catalog_name=from_catalog.name)
        from_table_1 = make_table(catalog_name=from_catalog.name, schema_name=from_schema.name)
        logger.info(f"Created new schema: {from_table_1}")
    ```
    """

    def create(
        *, catalog_name: str = "hive_metastore", name: str | None = None, location: str | None = None
    ) -> SchemaInfo:
        name = name or f"dummy_s{make_random(4)}".lower()
        full_name = f"{catalog_name}.{name}".lower()
        schema_ddl = f"CREATE SCHEMA {full_name}"
        if location:
            schema_ddl = f"{schema_ddl} LOCATION '{location}'"
        schema_ddl = f"{schema_ddl} WITH DBPROPERTIES (RemoveAfter={watchdog_remove_after})"
        sql_backend.execute(schema_ddl)
        schema_info = SchemaInfo(catalog_name=catalog_name, name=name, full_name=full_name, storage_location=location)
        path = f'explore/data/{schema_info.catalog_name}/{schema_info.name}'
        log_workspace_link(f'{schema_info.full_name} schema', path)
        return schema_info

    def remove(schema_info: SchemaInfo):
        sql_backend.execute(f"DROP SCHEMA IF EXISTS {schema_info.full_name} CASCADE")

    yield from factory("schema", create, remove)


@fixture
def make_catalog(
    ws, make_random, watchdog_remove_after, log_workspace_link
) -> Generator[Callable[..., CatalogInfo], None, None]:
    """
    Create a catalog and return its info. Remove it after the test.
    Returns instance of `databricks.sdk.service.catalog.CatalogInfo`.

    Keyword Arguments:
    * `name` (str): The name of the catalog. Default is a random string.

    Usage:
    ```python
    def test_catalog_fixture(make_catalog, make_schema, make_table):
        from_catalog = make_catalog()
        from_schema = make_schema(catalog_name=from_catalog.name)
        from_table_1 = make_table(catalog_name=from_catalog.name, schema_name=from_schema.name)
        logger.info(f"Created new schema: {from_table_1}")
    ```
    """

    def create(*, name: str | None = None) -> CatalogInfo:
        name = name or f"dummy_c{make_random(4)}".lower()
        catalog_info = ws.catalogs.create(name=name, properties={"RemoveAfter": watchdog_remove_after})
        if isinstance(catalog_info, Mock):
            catalog_info.name = name
        log_workspace_link(f'{name} catalog', f'explore/data/{name}')
        return catalog_info

    yield from factory(
        "catalog",
        create,
        lambda catalog_info: ws.catalogs.delete(catalog_info.full_name, force=True),
    )


@fixture
def make_udf(
    ws,
    env_or_skip,
    sql_backend,
    make_schema,
    make_random,
) -> Generator[Callable[..., FunctionInfo], None, None]:
    """
    Create a UDF and return its info. Remove it after the test. Returns instance of `databricks.sdk.service.catalog.FunctionInfo`.

    Keyword Arguments:
    * `catalog_name` (str): The name of the catalog where the UDF will be created. Default is `hive_metastore`.
    * `schema_name` (str): The name of the schema where the UDF will be created. Default is a random string.
    * `name` (str): The name of the UDF. Default is a random string.
    * `hive_udf` (bool): If `True`, the UDF will be created as a Hive UDF. Default is `False`.

    Usage:
    ```python
    def test_make_some_udfs(make_schema, make_udf):
        schema_a = make_schema(catalog_name="hive_metastore")
        make_udf(schema_name=schema_a.name)
        make_udf(schema_name=schema_a.name, hive_udf=True)
    ```
    """

    def create(
        *,
        catalog_name="hive_metastore",
        schema_name: str | None = None,
        name: str | None = None,
        hive_udf: bool = False,
    ) -> FunctionInfo:
        if schema_name is None:
            schema = make_schema(catalog_name=catalog_name)
            catalog_name = schema.catalog_name
            schema_name = schema.name

        name = name or f"dummy_f{make_random(4)}".lower()

        # Note: the Watchdog does not explicitly scan for functions; they are purged along with their parent schema.
        # As such the function can't be marked (and doesn't need to be if the schema as marked) for purge protection.

        full_name = f"{catalog_name}.{schema_name}.{name}".lower()
        if hive_udf:
            cmd_exec = CommandExecutor(
                ws.clusters,
                ws.command_execution,
                lambda: env_or_skip("TEST_DEFAULT_CLUSTER_ID"),
                language=Language.SQL,
            )
            ddl = f"CREATE FUNCTION {full_name} AS 'org.apache.hadoop.hive.ql.udf.generic.GenericUDFAbs';"
            cmd_exec.run(ddl)
        else:
            ddl = f"CREATE FUNCTION {full_name}(x INT) RETURNS FLOAT CONTAINS SQL DETERMINISTIC RETURN 0;"
            sql_backend.execute(ddl)
        udf_info = FunctionInfo(
            catalog_name=catalog_name,
            schema_name=schema_name,
            name=name,
            full_name=full_name,
        )

        logger.info(f"Function {udf_info.full_name} created")
        return udf_info

    def remove(udf_info: FunctionInfo):
        sql_backend.execute(f"DROP FUNCTION IF EXISTS {udf_info.full_name}")

    yield from factory("table", create, remove)


@fixture
def make_storage_credential(ws, watchdog_remove_after) -> Generator[Callable[..., StorageCredentialInfo], None, None]:
    """
    Create a storage credential and return its info. Remove it after the test. Returns instance of `databricks.sdk.service.catalog.StorageCredentialInfo`.

    Keyword Arguments:
    * `credential_name` (str): The name of the storage credential. Default is a random string.
    * `application_id` (str): The application ID for the Azure service principal. Default is an empty string.
    * `client_secret` (str): The client secret for the Azure service principal. Default is an empty string.
    * `directory_id` (str): The directory ID for the Azure service principal. Default is an empty string.
    * `aws_iam_role_arn` (str): The ARN of the AWS IAM role. Default is an empty string.
    * `read_only` (bool): If `True`, the storage credential will be read-only. Default is `False`.

    Usage:
    ```python
    def test_storage_credential(env_or_skip, make_storage_credential, make_random):
        random = make_random(6).lower()
        credential_name = f"dummy-{random}"
        make_storage_credential(
            credential_name=credential_name,
            aws_iam_role_arn=env_or_skip("TEST_UBER_ROLE_ID"),
        )
    ```
    """

    def create(
        *,
        credential_name: str,
        application_id: str = "",
        client_secret: str = "",
        directory_id: str = "",
        aws_iam_role_arn: str = "",
        read_only=False,
    ) -> StorageCredentialInfo:
        comment = json.dumps({"RemoveAfter": watchdog_remove_after})
        if aws_iam_role_arn != "":
            storage_credential = ws.storage_credentials.create(
                credential_name,
                aws_iam_role=AwsIamRoleRequest(role_arn=aws_iam_role_arn),
                read_only=read_only,
                comment=comment,
            )
        else:
            azure_service_principal = AzureServicePrincipal(directory_id, application_id, client_secret)
            storage_credential = ws.storage_credentials.create(
                credential_name, azure_service_principal=azure_service_principal, read_only=read_only, comment=comment
            )
        return storage_credential

    def remove(storage_credential: StorageCredentialInfo):
        ws.storage_credentials.delete(storage_credential.name, force=True)

    yield from factory("storage_credential", create, remove)


@fixture
def make_volume(
    ws, make_catalog, make_schema, make_random, log_workspace_link
) -> Generator[Callable[..., VolumeInfo], None, None]:
    """
    Create a volume and return its info. Remove it after the test. Returns instance of `databricks.sdk.service.catalog.VolumeInfo`.

    Keyword Arguments:
    * `catalog_name` (str): The name of the catalog where the schema and the volume are.
    * `schema_name` (str): The name of the schema where the volume is.
    * `name` (str): The name of the volume.
    * `comment` (str, optional): The comment attached to the volume.

    Usage:
    ```python
    def test_volume_creation(make_catalog, make_schema, make_volume, make_random):
        # Create a catalog
        catalog = make_catalog()

        # Create a schema in the catalog
        schema = make_schema(catalog_name=catalog.name)

        # Generate a random name for the volume
        volume_name = f"dummy_vol_{make_random(6).lower()}"

        # Create the volume
        volume = make_volume(
            catalog_name=catalog.name,
            schema_name=schema.name,
            name=volume_name
        )
    ```
    """

    def create(
        *,
        catalog_name: str | None = None,
        schema_name: str | None = None,
        name: str | None = None,
    ) -> VolumeInfo:

        if not catalog_name:
            catalog = make_catalog()
            catalog_name = catalog.name

        if not schema_name:
            schema = make_schema(catalog_name=catalog_name)
            schema_name = schema.name

        if not name:
            name = f"dummy_v{make_random(6).lower()}"

        volume_info = ws.volumes.create(
            catalog_name=catalog_name,
            schema_name=schema_name,
            name=name,
            volume_type=VolumeType.MANAGED,
        )
        path = f'explore/data/{volume_info.catalog_name}/{volume_info.schema_name}/{volume_info.name}'
        log_workspace_link(f'{volume_info.name} volume', path)
        return volume_info

    def remove(volume_info: VolumeInfo):
        ws.volumes.delete(f"{volume_info.catalog_name}.{volume_info.schema_name}.{volume_info.name}")

    yield from factory("volume", create, remove)



from functools import partial

from pytest import fixture
from databricks.labs.lsql.backends import StatementExecutionBackend


@fixture
def sql_backend(ws, env_or_skip) -> StatementExecutionBackend:
    """
    Create and provide a SQL backend for executing statements.

    Requires the environment variable `DATABRICKS_WAREHOUSE_ID` to be set.
    """
    warehouse_id = env_or_skip("DATABRICKS_WAREHOUSE_ID")
    return StatementExecutionBackend(ws, warehouse_id)


@fixture
def sql_exec(sql_backend):
    """
    Execute SQL statement and don't return any results.
    """
    return partial(sql_backend.execute)


@fixture
def sql_fetch_all(sql_backend):
    """
    Fetch all rows from a SQL statement.
    """
    return partial(sql_backend.fetch)



from datetime import timedelta, datetime, timezone
from pytest import fixture

TEST_RESOURCE_PURGE_TIMEOUT = timedelta(hours=1)


@fixture
def watchdog_remove_after() -> str:
    """
    Purge time for test objects, representing the (UTC-based) hour from which objects may be purged.
    """
    # Note: this code is duplicated in the workflow installer (WorkflowsDeployment) so that it can avoid the
    # transitive pytest deployment from this module.
    now = datetime.now(timezone.utc)
    purge_deadline = now + TEST_RESOURCE_PURGE_TIMEOUT
    # Round UP to the next hour boundary: that is when resources will be deleted.
    purge_hour = purge_deadline + (datetime.min.replace(tzinfo=timezone.utc) - purge_deadline) % timedelta(hours=1)
    return purge_hour.strftime("%Y%m%d%H")


@fixture
def watchdog_purge_suffix(watchdog_remove_after) -> str:
    """
    HEX-encoded purge time suffix for test objects.
    """
    return f'ra{int(watchdog_remove_after):x}'



"""Unwrapping pytest fixtures for unit testing."""

import inspect
from collections.abc import Callable, Generator
from typing import TypeVar
from unittest.mock import MagicMock, create_autospec

import pytest

from databricks.labs.lsql.backends import MockBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.service import iam

import databricks.labs.pytester.fixtures.plugin as P

T = TypeVar('T')


# Pytest fixtures are not supposed to be called directly, but historically we hack through the countermeasures.
# TODO: Investigate and fix this if possible, to avoid breakage in future pytest versions.
# Potential solution: use `pytest.FixtureRequest` & `request.getfixturevalue()` to access fixtures.
if pytest.version_tuple >= (8, 4):

    def call_fixture(fixture_fn: Callable[..., T], *args, **kwargs) -> T:
        if not hasattr(fixture_fn, "_get_wrapped_function"):
            raise ValueError(f'{fixture_fn} is not a pytest fixture')
        accessor = getattr(fixture_fn, "_get_wrapped_function")
        wrapped = accessor()
        return wrapped(*args, **kwargs)

else:
    # Older versions of pytest use a different mechanism to wrap fixtures.
    def call_fixture(fixture_fn: Callable[..., T], *args, **kwargs) -> T:
        if not hasattr(fixture_fn, '__pytest_wrapped__'):
            raise ValueError(f'{fixture_fn} is not a pytest fixture')
        wrapped = getattr(fixture_fn, '__pytest_wrapped__')
        if not hasattr(wrapped, 'obj'):
            raise ValueError(f'{fixture_fn} is not a pytest fixture')
        return wrapped.obj(*args, **kwargs)


class CallContext:
    def __init__(self):
        workspace_client = create_autospec(WorkspaceClient)
        workspace_client.current_user.me.return_value = iam.User(
            user_name="test-user",
            groups=[iam.ComplexValue(display="admins")],
        )
        workspace_client.config.host = "https://adb-12345679.10.azuredatabricks.net/"
        workspace_client.get_workspace_id.return_value = 123456789
        self._fixtures = {
            'sql_backend': MockBackend(),
            'ws': workspace_client,
            'make_random': self.make_random,
            'env_or_skip': self.env_or_skip,
            'watchdog_remove_after': '2024091313',
            'watchdog_purge_suffix': 'XXXXX',
        }

    def __getitem__(self, name: str):
        if name in self._fixtures:
            return self._fixtures[name]
        self._fixtures[name] = MagicMock()  # pylint: disable=obscure-mock
        return self._fixtures[name]

    def or_mock(self, name: str):
        if name in self._fixtures:
            return self._fixtures[name]
        self._fixtures[name] = MagicMock()  # pylint: disable=obscure-mock
        return self._fixtures[name]

    def __setitem__(self, key, value):
        self._fixtures[key] = value

    def __contains__(self, item):
        return item in self._fixtures

    @staticmethod
    def make_random(_=None):
        return 'RANDOM'

    @staticmethod
    def env_or_skip(name: str) -> str:
        return name

    def __str__(self):
        names = ', '.join(self._fixtures.keys())
        return f'CallContext<{names}>'


_GENERATORS = set[str]()


def call_stateful(
    some: Callable[..., Generator[Callable[..., T]]],
    call_context_setup: Callable[[CallContext], CallContext] = lambda x: x,
    **kwargs,
) -> tuple[CallContext, T]:
    # pylint: disable=too-complex
    ctx = CallContext()
    ctx = call_context_setup(ctx)
    drains = []

    if len(_GENERATORS) == 0:
        # discover all generator fixtures
        for name in P.__all__:
            fixture_fn = getattr(P, name)
            sig = inspect.signature(fixture_fn)
            if not hasattr(sig.return_annotation, '__origin__'):
                continue
            if sig.return_annotation.__origin__ != Generator:
                continue
            _GENERATORS.add(name)

    def _bfs_call_context(fn: Callable) -> Generator:
        sig = inspect.signature(fn)
        init = {}
        for param in sig.parameters.values():
            if param.name in _GENERATORS:
                upstream_fixture = getattr(P, param.name)
                init[param.name] = _bfs_call_context(upstream_fixture)
                continue
            init[param.name] = ctx.or_mock(param.name)
        yielding = call_fixture(fn, **init)
        ctx[fn.__name__] = next(yielding)
        drains.append(yielding)
        return ctx[fn.__name__]

    _bfs_call_context(some)
    result = ctx[some.__name__](**kwargs)

    for generator in reversed(drains):
        try:  # drain the generator and call cleanup
            next(generator)
        except StopIteration:
            pass

    return ctx, result
